unit uBTree;
{
    This file is part of Dizzy.

    Dizzy is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Dizzy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Dizzy; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
// This is a generic b-tree. It can be used to store any kind of objects,
// as long as a getValue Function is provided. Typecast the results with "as".
// It can be used as a (sorted) linked list, too.

interface

type
  // Prototype for getValue function of objects.
  TValueFunction = function (o : TObject; c : TClass): Integer of object;
  // Pointer for tree branches.
  TTreeBranchPointer = ^TTreeBranch;
  // The actual tree structure.
  TTreeBranch = record
    Content : TObject;
    Less, More, Parent: TTreeBranchPointer;
  end;
  // The interface to the b-tree.
  TTree = class
    public
      constructor Create(c : TClass; vf : TValueFunction);
      destructor Destroy; override;
      procedure recursiveDispose(p : TTreeBranchPointer);
      // These are the essential prcedures.
      procedure insertItem(o : TObject);
      procedure deleteItem;
      procedure recursiveReenter(p : TTreeBranchPointer);
      // These are the essential functions.
      function getItemClass : TClass;
      function resetPosition : TObject;
      function getCurrent : TObject;
      function getNext : TObject;
      function getPrev : TObject;
      // Use these for quick navigation.
      function getNearFloorValue(v : Integer) : TObject;
      function getNearCeilValue(v : Integer) : TObject;
      function getByValue(v : Integer) : TObject;
      function getNextByValue(v : Integer) : TObject;
      // These functions return pointers. Beware.
      function getLess : TTreeBranchPointer;
      function getMore : TTreeBranchPointer;
      function getParent : TTreeBranchPointer;
      // You can use this one to meddle with the "Current" pointer. Beware.
      procedure setCurrent (p : TTreeBranchPointer);
    private
      ItemClass : TClass;
      ValueFunction : TValueFunction;
      Root, Current : TTreeBranchPointer;
  end;

implementation

constructor TTree.Create(c : TClass; vf : TValueFunction);
begin
  ItemClass := c;
  ValueFunction := vf;
  Root := nil;
  Current := nil;
end;

destructor TTree.Destroy;
begin
  recursiveDispose(Root);
  inherited Destroy;
end;

procedure TTree.recursiveDispose(p : TTreeBranchPointer);
begin
  if p = nil then exit;
  if p^.Less <> nil then recursiveDispose(p^.Less);
  if p^.More <> nil then recursiveDispose(p^.More);
  p^.Content.Free;
  //dispose(p);
end;

// Essential procedures.
//----------------------------------------------------------------------

// Remember to resetPosition after inserting item.
procedure TTree.insertItem (o : TObject);
var
  Iterator, Parent : TTreeBranchPointer;
begin
  // Check the tree's root.
  if Root = nil then
    begin
      new(Root);
      Root^.Content := o;
      Root^.Less := nil;
      Root^.More := nil;
      Root^.Parent := nil;
      exit;
    end;
  // Find an empty branch.
  Iterator := Root;
  Parent := Iterator;
  while Iterator <> nil do
    begin
      if ValueFunction(o, ItemClass) > ValueFunction(Iterator^.Content, ItemClass) then
        begin
          Parent := Iterator;
          Iterator := Iterator^.More;
        end
      else
        begin
          Parent := Iterator;
          Iterator := Iterator^.Less;
        end;
    end;
  // Create branch.
  new(Iterator);
  Iterator^.Content := o;
  Iterator^.Less := nil;
  Iterator^.More := nil;
  Iterator^.Parent := Parent;
  // Glue branch to tree.
  if ValueFunction(o, ItemClass) > ValueFunction(Parent^.Content, ItemClass) then
    Parent^.More := Iterator
  else
    Parent^.Less := Iterator;
end;

procedure TTree.recursiveReenter(p : TTreeBranchPointer);
begin
  if p = nil then exit;
  insertItem(p^.Content);
  recursiveReenter(p^.More);
  recursiveReenter(p^.Less);
  dispose(p);
end;

procedure TTree.deleteItem;
begin
  if Current = nil then exit;
  if Current^.Parent = nil then
    Root := nil
  else
    if Current = Current^.Parent^.Less then
      Current^.Parent^.Less := nil
    else
      Current^.Parent^.More := nil;
  recursiveReenter(Current^.More);
  recursiveReenter(Current^.Less);
  dispose(Current);
  Current := Root;
end;

// Essential functions.
//----------------------------------------------------------------------

function TTree.getItemClass : TClass;
begin
  result := ItemClass;
end;

function TTree.resetPosition : TObject;
begin
  result := nil;
  Current := Root;
  if Root = nil then exit;
  // If there is anything in the tree, fall down the less-branches.
  while Current^.Less <> nil do
    Current := Current^.Less;
  result := Current^.Content;
end;

function TTree.getCurrent : TObject;
begin
  // First check if we are anywhere at all.
  if Current = nil then
    begin
      result := nil;
      exit;
    end;
  result := Current^.Content;
end;

function TTree.getNext : TObject;
begin
  // First check if we are anywhere at all.
  if Current = nil then
    begin
      result := nil;
      exit;
    end;
  // Are we on a less-branch or at the top?
  if (Current^.Parent = nil) or (Current = Current^.Parent^.Less) or (Current^.More <> nil) then
  begin
    // Can we go deeper?
    if Current^.More = nil then
      begin
        // No, do we have a parent?
        if Current^.Parent = nil then
          begin
            // Void.
            Current := nil;
            result := nil;
            exit;
          end;
        // We'll become our parent.
        Current := Current^.Parent;
        result := Current^.Content;
        exit;
      end;
    // We can go deeper. Go to the more-branch and then fall down the less-branches.
    Current := Current^.More;
    while Current^.Less <> nil do
      Current := Current^.Less;
    result := Current^.Content;
    exit;
  end;
  // We are a more-branch.
  while (Current^.Parent <> nil) and (Current <> Current^.Parent^.Less) do
    Current := Current^.Parent;
  Current := Current^.Parent;
  // Are we done?
  if Current = nil then
    begin
      // Yes, we are. Return nil.
      result := nil;
      exit;
    end;
  result := Current^.Content;
end;

function TTree.getPrev : TObject;
begin
  // First check if we are anywhere at all.
  if (Current = nil) or ((Current^.Less = nil) and (Current^.Parent = nil)) then
    begin
      result := nil;
      exit;
    end;
  // Is there a less-branch?
  if Current^.Less <> nil then
    begin
      // Yes, there is. Go there and find the highest value.
      Current := Current^.Less;
      while Current^.More <> nil do
        Current := Current^.More;
      result := Current^.Content;
      exit;
    end;
  // No less-branch and we are on a left-branch.
  if (Current^.Less = nil) and (Current^.Parent^.Less = Current) then
    begin
      while (Current^.Parent <> nil) and (Current^.Parent^.Less = Current) do
        Current := Current^.Parent;
    end;
  // There is no less-branch. Go up.
  Current := Current^.Parent;
  // Is there still something left?
  if Current = nil then
    begin
      // There is nothing left for us to retrieve.
      result := nil;
      exit;
    end;
  // Yes, there is.
  result := Current^.Content;
end;

// Stuff for quick navigation.
//----------------------------------------------------------------------

// Returns the last item before the value reaches v.
function TTree.getNearFloorValue(v : Integer) : TObject;
var
  Last : TTreeBranchPointer;
begin
  Last := nil;
  resetPosition;
  while (Current <> nil) and (v > ValueFunction(Current^.Content, ItemClass)) do
    begin
      Last := Current;
      getNext;
    end;
  Current := Last;
  if Current <> nil then result := Current^.Content else result := nil;
end;

// Returns the next bigger value than v.
function TTree.getNearCeilValue(v : Integer) : TObject;
begin
  resetPosition;
  while (Current <> nil) and (v <= ValueFunction(Current^.Content, ItemClass)) do
    begin
      getNext;
    end;
  if Current <> nil then result := Current^.Content else result := nil;
end;

function TTree.getByValue(v : Integer) : TObject;
begin
  Current := Root;
  while (Current <> nil) and (v <> ValueFunction(Current^.Content, ItemClass)) do
    begin
      if v > ValueFunction(Current^.Content, ItemClass) then
        Current := Current^.More
      else
        Current := Current^.Less;
    end;
  if Current <> nil then result := Current^.Content else result := nil;
end;

// Use only after having found v with getByValue.
function TTree.getNextByValue(v : Integer) : TObject;
begin
  if v = ValueFunction(Current^.Less^.Content, ItemClass) then
    begin
      Current := Current^.Less;
      if Current <> nil then result := Current^.Content else result := nil;
    end
  else
    begin
      Current := nil;
      result := nil;
    end;
end;

// The following ought to be used with care. Raw data handling.
//----------------------------------------------------------------------

function TTree.getLess : TTreeBranchPointer;
begin
  if Current^.Less = nil then
    begin
      result := nil;
      exit;
    end;
  result := Current^.Less;
end;

function TTree.getMore : TTreeBranchPointer;
begin
  if Current^.More = nil then
    begin
      result := nil;
      exit;
    end;
  result := Current^.More;
end;

function TTree.getParent : TTreeBranchPointer;
begin
  if Current^.Parent = nil then
    begin
      result := nil;
      exit;
    end;
  result := Current^.Parent;
end;

procedure TTree.setCurrent (p : TTreeBranchPointer);
begin
  Current := p;
end;

end.
