unit uGate;
{
    This file is part of Dizzy.

    Dizzy is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Dizzy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Dizzy; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}

interface

uses uBTree, uMovableObject, Classes, Controls, Graphics, Windows, SysUtils;

type
  // Forward declaration of TGate.
  TGate = class;
  //------------------------------------------------------------------------
  TPowerLine = class
    public
      constructor Create(ns : Integer; s : TGate; nd : Integer; d : TGate; c : TCanvas; i : Integer; q, p : TTree);
      constructor softCreate(ns : Integer; s : TGate; nd : Integer; d : TGate; v : Boolean; c : TCanvas; i : Integer; q, p : TTree);
      destructor Destroy; override;
      // This procedure is used to change the value of the powerline.
      procedure changeQuick(q : Boolean);
      procedure setValue(v : Boolean);
      function getValue : Boolean;
      // For other's value calculation.
      function getId : Integer;
      // Drawing procedure.
      procedure blitToCanvas;
      // Used for saving.
      function getDataString : String;
    private
      S, D : TGate;
      NS, ND, Id : Integer;
      targetCanvas : TCanvas;
      Queue, PowerLines : TTree;
      Value, Quick : Boolean;
  end;
  //------------------------------------------------------------------------
  TClickSections = record
    Input, Output : Integer;
  end;
  //------------------------------------------------------------------------
  TGate = class(TMob)
    public
      constructor Create(x, y, i, o : Integer; c : TCanvas); override;
      destructor Destroy; override;
      procedure blitToCanvas; override;
      // Used for loading.
      constructor loadFromString(s : String; c : TCanvas); override;
      // This procedures makes the gate recalculate everything.
      procedure signalChanged;
      procedure reloadValues; virtual;
      procedure recalculateValues; virtual;
      // Functions for setting input sources and output destinations etc..
      procedure setIA(n : Integer; i : TPowerLine); virtual; abstract;
      procedure setOA(n : Integer; o : TTree); virtual; abstract;
      procedure notifyOutputJack(n : Integer; v : Boolean);
      procedure destroyConnection(n : Integer);
      function emptyJack(n : Integer) : Boolean;
      // Functions that help "TPowerLine" find it's source and target.
      function getInputCoords(n : Integer) : TPoint;
      function getOutputCoords(n : Integer) : TPoint;
      // Event handler and section detection.
      function getClickedSection(x, y : Integer) : TClickSections;
      // Value calculation of TPowerLine.
      function getPLValue(o : TObject; c : TClass) : Integer;
      // For other's value calculation.
      function getId : Integer;
      // Used for saving.
      function getDataString : String; virtual;
    protected
      I, O, Id : Integer;
      // Drawing input and output jacks to "ownPicture".
      procedure drawJacks;
      // Abstractors for array access.
      procedure setVA(n : Integer; v : Boolean); virtual; abstract;
      function getIA(n : Integer) : TPowerLine; virtual; abstract;
      function getOA(n : Integer) : TTree; virtual; abstract;
  end;

implementation

constructor TGate.Create(x, y, i, o : Integer; c : TCanvas);
var
  j : Integer;
begin
  inherited Create(x, y, 50, 120, c);
  self.I := i;
  self.O := o;
  drawJacks;
  // Set start values.
  for j := 1 to I do
    setVA(j, true);
  for j := 1 to I do
    setIA(j, nil);
  for j := 1 to O do
    setOA(j, TTree.Create(TPowerLine, getPLValue));
end;

destructor TGate.Destroy;
var
  j : Integer;
begin
  for j := 1 to I do
    if getIA(j) <> nil then getIA(j).Destroy;
  for j := 1 to O do
      getOA(j).Destroy;
  inherited Destroy;
end;

procedure TGate.blitToCanvas;
var
  j : Integer;
begin
  inherited blitToCanvas;
  for j := 1 to I do
    if getIA(j) <> nil then getIA(j).blitToCanvas;
end;

constructor TGate.loadFromString(s : String; c : TCanvas);
var
  id, x, y, EndOfSection : Integer;
begin
  id := StrToInt(Copy(s, 1, Pos(':', s) - 1));
  Delete(s, 1, Pos(':', s));
  x := StrToInt(Copy(s, 1, Pos(':', s) - 1));
  Delete(s, 1, Pos(':', s));
  if Pos(':', s) > 0 then
    EndOfSection := Pos(':', s)
  else
    begin
      s := s + ':';
      EndOfSection := Length(s);
    end;
  y := StrToInt(Copy(s, 1, EndOfSection - 1));
  Delete(s, 1, EndOfSection);
  Init(x, y, c, id);
end;

procedure TGate.reloadValues;
var
  j : Integer;
begin
  for j := 1 to I do
    if getIA(j) <> nil then
      setVA(j, getIA(j).getValue)
    else
      setVA(j, true);
end;

procedure TGate.recalculateValues;
begin
end;

procedure TGate.signalChanged;
begin
  reloadValues;
  recalculateValues;
end;

procedure TGate.notifyOutputJack(n : Integer; v : Boolean);
var
  PowerLine : TPowerLine;
begin
  PowerLine := getOA(n).resetPosition as TPowerLine;
  while PowerLine <> nil do
    begin
      PowerLine.setValue(v);
      PowerLine := getOA(n).getNext as TPowerLine;
    end;
end;


procedure TGate.destroyConnection(n : Integer);
begin
  getIA(n).Destroy;
end;

function TGate.emptyJack(n : Integer) : Boolean;
begin
  if getIA(n) = nil then
    result := true
  else
    result := false;
end;

procedure TGate.drawJacks;
var
  J, JackHeight : Integer;
begin
  with ownPicture.Canvas do
    begin
      // Draw the box.
      Pen.Color := clBlack;
      Brush.Color := clLtGray;
      Rectangle(0, 0, W, H);
      // Any input jacks?
      if I > 0 then
        begin
          // Draw a separator.
          MoveTo(10, 0);
          LineTo(10, H);
          // Find out, how big each jack has to be.
          JackHeight := H div I;
          for J := 1 to I do
            begin
              // Draw the separators between the jacks.
              MoveTo(0, J * JackHeight);
              LineTo(10, J * JackHeight);
            end;
        end;
      // Any output jacks?
      if O > 0 then
        begin
          // Draw a separator.
          MoveTo(W - 10, 0);
          LineTo(W - 10, H);
          // Find out, how big each jack has to be.
          JackHeight := H div O;
          for J := 1 to I do
            begin
              // Draw the separators between the jacks.
              MoveTo(W - 10, J * JackHeight);
              LineTo(W, J * JackHeight);
            end;
        end;
    end;
end;

function TGate.getInputCoords(n : Integer) : TPoint;
var
  JackCoords : TPoint;
begin
  JackCoords.X := X + 5;
  JackCoords.Y := Y + (n * (H div I)) - ((H div I) div 2);
  result := JackCoords;
end;

function TGate.getOutputCoords(n : Integer) : TPoint;
var
  JackCoords : TPoint;
begin
  JackCoords.X := X + W - 5;
  JackCoords.Y := Y + (n * (H div O)) - ((H div O) div 2);
  result := JackCoords;
end;

function TGate.getClickedSection(x, y : Integer) : TClickSections;
var
  Structure : TClickSections;
begin
  x := x - self.X;
  y := y - self.Y;
  Structure.Input := 0;
  Structure.Output := 0;
  if I > 0 then
    if x <= 10 then
      Structure.Input := 1 + Y div (H div I);
  if O > 0 then
    if x >= W - 10 then
      Structure.Output := 1 + Y div (H div O);
  result := Structure;
end;

function TGate.getId : Integer;
begin
  result := Id;
end;

function TGate.getPLValue(o : TObject; c : TClass) : Integer;
var
  PL : TPowerLine;
begin
  PL := o as TPowerLine;
  result := PL.getId;
end;

function TGate.getDataString : String;
var
  DataString, PLString : String;
  j : Integer;
begin
  PLString := '';
  DataString := ':' + IntToStr(getId) + ':' + IntToStr(X) + ':' + IntToStr(Y);
  for j := 1 to I do
    if getIA(j) <> nil then PLString := PLString + ';' + getIA(j).getDataString;
  if PLString <> '' then
    begin
      Delete(PLString, 1, 1);
      DataString := DataString + '|' + PLString;
    end;
  result := DataString;
end;

constructor TPowerLine.Create(ns : Integer; s : TGate; nd : Integer; d : TGate; c : TCanvas; i : Integer; q, p : TTree);
begin
  self.S := s;
  self.D := d;
  self.NS := ns;
  self.ND := ND;
  self.Id := i;
  self.Queue := q;
  self.PowerLines := p;
  self.Quick := true;
  targetCanvas := c;
  s.getOA(ns).insertItem(self);
  d.setIA(nd, self);
  // Make the source recalculate everything, so it can give us a signal to send.
  s.signalChanged;
  d.signalChanged;
end;

constructor TPowerLine.softCreate(ns : Integer; s : TGate; nd : Integer; d : TGate; v : Boolean; c : TCanvas; i : Integer; q, p : TTree);
begin
  self.S := s;
  self.D := d;
  self.NS := ns;
  self.ND := ND;
  self.Value := v;
  self.Id := i;
  self.Queue := q;
  self.PowerLines := p;
  self.Quick := true;
  targetCanvas := c;
  s.getOA(ns).insertItem(self);
  d.setIA(nd, self);
  blitToCanvas;
end;

procedure TPowerLine.setValue(v : Boolean);
begin
  if v = Value then exit;
  Value := v;
  blitToCanvas;
  if Quick then
    begin
      D.signalChanged;
      D.blitToCanvas;
      exit;
    end;
  if (Queue.getCurrent as TTree).getByValue(D.getId) = nil then (Queue.getCurrent as TTree).insertItem(D);
end;

function TPowerLine.getValue : Boolean;
begin
  result := Value;
end;

procedure TPowerLine.blitToCanvas;
var
  OutputJackCoords, InputJackCoords : TPoint;
begin
  with targetCanvas do
    begin
      if Value then
        Pen.Color := clRed
      else
        Pen.Color := clBlue;
      OutputJackCoords := S.getOutputCoords(NS);
      InputJackCoords := D.getInputCoords(ND);
      MoveTo(OutputJackCoords.X, OutputJackCoords.Y);
      LineTo(InputJackCoords.X, InputJackCoords.Y);
    end;
end;

procedure TPowerLine.changeQuick(q : Boolean);
begin
  Quick := q;
end;

destructor TPowerLine.Destroy;
begin
  S.getOA(ns).resetPosition;
  S.getOA(ns).getByValue(Id);
  S.getOA(ns).deleteItem;
  PowerLines.getByValue(Id);
  PowerLines.deleteItem;
  D.setIA(nd, nil);
  D.signalChanged;
  inherited Destroy;
end;

function TPowerLine.getId : Integer;
begin
  result := Id;
end;

function TPowerLine.getDataString : String;
var
  IVal : Integer;
begin
  if Value then
    IVal := 1
  else
    IVal := 0;
  result := IntToStr(S.getId) + ':' + IntToStr(NS) + ':' + IntToStr(D.getId) + ':' + IntToStr(ND) + ':' + IntToStr(IVal);
end;

end.
